from typing import Optional
from dataclasses import dataclass, field


@dataclass
class Credentials:
    username: str
    password: str
    enable: Optional[str] = field(default=None)

    def __post_init__(self):
        if not self.enable:
            self.enable = self.password



import asyncio
from os import access
from typing import Callable, List, Set, Awaitable

from net_gsd.host import Host, FailedHost


class Task:
    task_queue: set = set()
    failed_hosts: list = []
    debug: bool = False
    result: dict = {}
    complete: bool = False

    def __init__(self, task: Callable, name: str, hosts: List[Host]) -> None:
        self.task: Callable = task
        self.name: str = name
        self.hosts: List[Host] = hosts

    async def _get_hosts(self):
        for host in self.hosts:
            yield host

    async def run_task(self, debug: bool) -> list:
        """Starts the task running"""
        self.debug = debug
        for host in self.hosts:
            if self.debug:
                print(f"running task for {host.hostname}")
            task = asyncio.get_event_loop().create_task(self.task(host), name=f"{self.name}:{host.hostname}:{host.ip}")
            self.task_queue.add(task)
            task.add_done_callback(self._handle_task_result)
        await self._monitor_task()

    async def _monitor_task(self):
        """Monitors the task and removes finished tasks from task list"""
        while len(self.task_queue):
            complete, _ = await asyncio.wait(self.task_queue, timeout=0.5)
            self.task_queue.difference_update(complete)
            if self.debug:
                print(len(self.task_queue))
        self.complete = True

    async def monitor(self):
        """calls the monitor task function"""
        await self._monitor_task()

    def _handle_task_result(self, task: asyncio.Task) -> None:
        """handles the asyncio task callback once task has finshed"""
        try:
            result = task.result()
            hostname = task.get_name().split(":")[1]
            if result and isinstance(result, dict):
                self.result.update({hostname: result})
        except asyncio.CancelledError:
            pass
        except Exception as e:
            task_name, device, ip = task.get_name().split(":")
            self.failed_hosts.append(FailedHost(device, ip, task_name, getattr(e, "message", str(e))))

    def get_failed_hosts(self):
        return self.failed_hosts



from typing import *
from time import perf_counter

from net_gsd.runner import Credentials
from net_gsd.host import Host
from net_gsd.task import Task
from net_gsd.runner.errors import NoHostsPresent, InvalidHostKeys, InvalidTask


class Runner:
    """Task runner"""

    tasks: Set[Task] = set()
    failed_hosts: dict = {}

    def __init__(
        self, username: str, password: str, enable: Optional[str] = None, hosts: List[Dict] = None, debug: bool = False
    ) -> None:
        """Initialise task runner. Requires username and password"""
        self.credentials: Credentials = Credentials(username=username, password=password, enable=enable)
        self.hosts: list[Host] = self.get_hosts(hosts) if hosts else []
        self.debug: bool = debug

    def get_hosts(self, hosts: List[Dict]) -> None:
        """Add hosts to host list"""
        try:
            hosts = [
                Host(hostname=host["hostname"], ip=host["ip"], platform=host["platform"], credentials=self.credentials)
                for host in hosts
            ]
        except KeyError:
            raise InvalidHostKeys("Invalid or missing keys in host list")
        return hosts

    def _validate_task(self, name: str, task: Callable, hosts: List[Dict]) -> bool:
        if not isinstance(task, Callable):
            raise InvalidTask(f"{name}: Task must be a function")
        if len(self.hosts) == 0 and not hosts:
            raise NoHostsPresent("No hosts have been passed to the runner.")
        return True

    def queue_task(self, name: str, task: Callable, hosts: List[Dict] = None) -> None:
        """Add task to pending task list, if no hosts are passed into the function,
        the class hosts will be used"""
        if self._validate_task(name, task, hosts):
            task_hosts = self.get_hosts(hosts) if hosts else self.hosts
            self.tasks.add(Task(name=name, task=task, hosts=task_hosts))

    async def run_task(self, name: str, task: Callable, hosts: List[Dict] = None) -> Tuple[dict, list]:
        """Runs task straight away, if no hosts are passed into the function,
        the class hosts will be used"""
        if self._validate_task(name, task, hosts):
            task_hosts = self.get_hosts(hosts) if hosts else self.hosts
            task: Task = Task(name=name, task=task, hosts=task_hosts)
            await task.run_task(self.debug)
        return task.result, task.failed_hosts

    async def _get_tasks(self):
        """Get tasks generator"""
        for task in self.tasks:
            yield task

    async def run(self) -> None:
        """run tasks in the pending tasks queue"""
        start = perf_counter()
        async for task in self._get_tasks():
            print(f"Starting task {task.name}")
            await task.run_task(self.debug)
        end = perf_counter()
        print(f"Tasks ran in {end - start:0.4f} seconds")

    def get_failed_hosts(self) -> List[set]:
        """Return a list if failed hosts from all tasks"""
        return [task.failed_hosts[0] for task in self.tasks if task.failed_hosts]

    def tasks_complete(self) -> bool:
        """Return a bool statiung whether the task have completed"""
        return sum(len(task.task_queue) for task in self.tasks) == 0

    def failed_hosts(self) -> bool:
        """Returns a bool indicating whether there were any failed hosts"""
        return len([task.failed_hosts for task in self.tasks]) > 0

    def get_results(self) -> dict:
        """get results for all tasks"""
        results = {}
        for task in self.tasks:
            results.update(task.result)
        return results



from typing import Union, Dict
from contextlib import asynccontextmanager

from scrapli.driver import AsyncDriver
from scrapli.driver.core import AsyncIOSXEDriver, AsyncNXOSDriver
from scrapli.response import Response

from net_gsd.connections.base import Base

PLATFORM = {"ios": AsyncIOSXEDriver, "nxos": AsyncNXOSDriver}


class SSH(Base):
    _con: AsyncDriver = None

    def __init__(self, host: str, username: str, password: str, platform: str, enable: str) -> None:
        self.host: str = host
        self.username: str = username
        self.password: str = password
        self.enable: str = enable
        self.platform: str = platform

        self.get_driver()

    def get_driver(self) -> None:
        self._con = PLATFORM.get(self.platform)(
            host=self.host,
            auth_username=self.username,
            auth_password=self.password,
            auth_secondary=self.enable,
            auth_strict_key=False,
            transport="asyncssh",
        )

    @asynccontextmanager
    async def get_connection(self):
        await self._con.open()
        yield self
        await self._con.close()

    async def send_command(self, cmd: str, parse: bool = True) -> Union[Dict, str]:
        result: Response = await self._con.send_command(cmd)
        if parse:
            return result.genie_parse_output()
        return result.result



from typing import Union
from dataclasses import dataclass
from ipaddress import IPv4Address, AddressValueError

from net_gsd.connections import SSH
from net_gsd.runner import Credentials
from net_gsd.host.errors import InvalidIPAddress, InvalidPlatform
from net_gsd.connections.ssh import PLATFORM


class Host:
    def __init__(self, hostname: str, ip: str, platform: str, credentials: Credentials) -> None:
        if platform not in PLATFORM.keys():
            raise InvalidPlatform(f"Host {hostname}: platform must be {list(PLATFORM.keys())}")
        self.hostname: str = hostname
        try:
            self.ip: str = str(IPv4Address(ip))
        except AddressValueError:
            raise InvalidIPAddress(f"Host: {hostname} - has an invalid ip address of {ip}")
        self.platform: str = platform
        self.connection: SSH = SSH(
            host=ip,
            platform=platform,
            username=credentials.username,
            password=credentials.password,
            enable=credentials.enable,
        )

    def __repr__(self):
        return f"Host(Hostname: {self.hostname}, IP: {self.ip})"

    async def send_command(self, cmds: list, parse: bool = True) -> Union[str, dict]:
        async with self.connection.get_connection() as con:
            results = {cmd: await con.send_command(cmd, parse) for cmd in cmds}
        return results


@dataclass
class FailedHost:
    hostname: str
    ip: IPv4Address
    task: str
    error: str



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['net_gsd',
 'net_gsd.connections',
 'net_gsd.host',
 'net_gsd.runner',
 'net_gsd.task']

package_data = \
{'': ['*']}

install_requires = \
['asyncssh>=2.7.0,<3.0.0',
 'genie==21.4',
 'pyats==21.4',
 'scrapli>=2021.1.30,<2022.0.0']

setup_kwargs = {
    'name': 'net-gsd',
    'version': '0.1.2',
    'description': 'Network Get Stuff Done',
    'long_description': None,
    'author': 'Ryan Bradshaw',
    'author_email': 'ryan@rbradshaw.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
